@@ -1,3 +1,5 @@
+#![cfg(not(tarpaulin_include))]
+
 use evalexpr::{error::*, *};
 
 #[test]
@@ -15,6 +17,11 @@ fn test_unary_examples() {
     assert_eq!(eval("-3"), Ok(Value::Int(-3)));
     assert_eq!(eval("-3.6"), Ok(Value::Float(-3.6)));
     assert_eq!(eval("----3"), Ok(Value::Int(3)));
+    assert_eq!(eval("1e0"), Ok(Value::Float(1.0)));
+    assert_eq!(eval("1e-0"), Ok(Value::Float(1.0)));
+    assert_eq!(eval("10e3"), Ok(Value::Float(10000.0)));
+    assert_eq!(eval("10e+3"), Ok(Value::Float(10000.0)));
+    assert_eq!(eval("10e-3"), Ok(Value::Float(0.01)));
 }
 
 #[test]
@@ -39,6 +46,8 @@ fn test_binary_examples() {
     assert_eq!(eval("3+-1"), Ok(Value::Int(2)));
     assert_eq!(eval("-3-5"), Ok(Value::Int(-8)));
     assert_eq!(eval("-5--3"), Ok(Value::Int(-2)));
+    assert_eq!(eval("5e2--3"), Ok(Value::Float(503.0)));
+    assert_eq!(eval("-5e-2--3"), Ok(Value::Float(2.95)));
 }
 
 #[test]
@@ -267,6 +276,42 @@ fn test_n_ary_functions() {
     );
 }
 
+#[test]
+fn test_capturing_functions() {
+    let mut context = HashMapContext::new();
+    // this variable is captured by the function
+    let three = 3;
+    context
+        .set_function(
+            "mult_3".into(),
+            Function::new(move |argument| {
+                if let Value::Int(int) = argument {
+                    Ok(Value::Int(int * three))
+                } else if let Value::Float(float) = argument {
+                    Ok(Value::Float(float * three as f64))
+                } else {
+                    Err(EvalexprError::expected_number(argument.clone()))
+                }
+            }),
+        )
+        .unwrap();
+
+    let four = 4;
+    context
+        .set_function(
+            "function_four".into(),
+            Function::new(move |_| Ok(Value::Int(four))),
+        )
+        .unwrap();
+
+    assert_eq!(eval_with_context("mult_3 2", &context), Ok(Value::Int(6)));
+    assert_eq!(eval_with_context("mult_3(3)", &context), Ok(Value::Int(9)));
+    assert_eq!(
+        eval_with_context("mult_3(function_four())", &context),
+        Ok(Value::Int(12))
+    );
+}
+
 #[test]
 fn test_builtin_functions() {
     // Log
@@ -274,6 +319,13 @@ fn test_builtin_functions() {
     assert_eq!(eval("math::log(9, 9)"), Ok(Value::Float(1.0)));
     assert_eq!(eval("math::log2(2)"), Ok(Value::Float(1.0)));
     assert_eq!(eval("math::log10(10)"), Ok(Value::Float(1.0)));
+    // Powers
+    assert_eq!(eval("math::exp(2)"), Ok(Value::Float(2.0_f64.exp())));
+    assert_eq!(eval("math::exp2(2)"), Ok(Value::Float(2.0_f64.exp2())));
+    assert_eq!(
+        eval("math::pow(1.5, 1.3)"),
+        Ok(Value::Float(1.5_f64.powf(1.3)))
+    );
     // Cos
     assert_eq!(eval("math::cos(0)"), Ok(Value::Float(1.0)));
     assert_eq!(eval("math::acos(1)"), Ok(Value::Float(0.0)));
@@ -289,9 +341,18 @@ fn test_builtin_functions() {
     assert_eq!(eval("math::atan(0)"), Ok(Value::Float(0.0)));
     assert_eq!(eval("math::tanh(0)"), Ok(Value::Float(0.0)));
     assert_eq!(eval("math::atanh(0)"), Ok(Value::Float(0.0)));
+    assert_eq!(
+        eval("math::atan2(1.2, -5.5)"),
+        Ok(Value::Float(1.2_f64.atan2(-5.5)))
+    );
     // Root
     assert_eq!(eval("math::sqrt(25)"), Ok(Value::Float(5.0)));
     assert_eq!(eval("math::cbrt(8)"), Ok(Value::Float(2.0)));
+    // Hypotenuse
+    assert_eq!(
+        eval("math::hypot(8.2, 1.1)"),
+        Ok(Value::Float(8.2_f64.hypot(1.1)))
+    );
     // Rounding
     assert_eq!(eval("floor(1.1)"), Ok(Value::Float(1.0)));
     assert_eq!(eval("floor(1.9)"), Ok(Value::Float(1.0)));
@@ -394,144 +455,673 @@ fn test_shortcut_functions() {
         .set_value("string".into(), Value::from("a string"))
         .unwrap();
 
-    // assert_eq!(eval_string("???"));
+    assert_eq!(eval_string("\"3.3\""), Ok("3.3".to_owned()));
+    assert_eq!(
+        eval_string("3.3"),
+        Err(EvalexprError::ExpectedString {
+            actual: Value::Float(3.3)
+        })
+    );
+    assert_eq!(
+        eval_string("3..3"),
+        Err(EvalexprError::VariableIdentifierNotFound("3..3".to_owned()))
+    );
     assert_eq!(
         eval_string_with_context("string", &context),
-        Ok("a string".to_string())
-    );
-    assert_eq!(eval_float("3.3"), Ok(3.3));
-    assert_eq!(eval_float_with_context("3.3", &context), Ok(3.3));
-    assert_eq!(eval_int("3"), Ok(3));
-    assert_eq!(eval_int_with_context("3", &context), Ok(3));
-    assert_eq!(eval_number("3"), Ok(3.0));
-    assert_eq!(eval_number_with_context("3", &context), Ok(3.0));
-    assert_eq!(eval_boolean("true"), Ok(true));
-    assert_eq!(eval_boolean_with_context("true", &context), Ok(true));
-    assert_eq!(eval_tuple("3,3"), Ok(vec![Value::Int(3), Value::Int(3)]));
-    assert_eq!(
-        eval_tuple_with_context("3,3", &context),
-        Ok(vec![Value::Int(3), Value::Int(3)])
-    );
-    assert_eq!(eval_empty(""), Ok(EMPTY_VALUE));
-    assert_eq!(eval_empty_with_context("", &context), Ok(EMPTY_VALUE));
-
-    // assert_eq!(build_operator_tree("???").unwrap().eval_string());
-    assert_eq!(
-        build_operator_tree("string")
-            .unwrap()
-            .eval_string_with_context(&context),
-        Ok("a string".to_string())
-    );
-    assert_eq!(build_operator_tree("3.3").unwrap().eval_float(), Ok(3.3));
-    assert_eq!(
-        build_operator_tree("3.3")
-            .unwrap()
-            .eval_float_with_context(&context),
-        Ok(3.3)
-    );
-    assert_eq!(build_operator_tree("3").unwrap().eval_int(), Ok(3));
-    assert_eq!(
-        build_operator_tree("3")
-            .unwrap()
-            .eval_int_with_context(&context),
-        Ok(3)
-    );
-    assert_eq!(build_operator_tree("3").unwrap().eval_number(), Ok(3.0));
-    assert_eq!(
-        build_operator_tree("3")
-            .unwrap()
-            .eval_number_with_context(&context),
-        Ok(3.0)
+        Ok("a string".to_owned())
     );
     assert_eq!(
-        build_operator_tree("true").unwrap().eval_boolean(),
-        Ok(true)
+        eval_string_with_context("3.3", &context),
+        Err(EvalexprError::ExpectedString {
+            actual: Value::Float(3.3)
+        })
     );
     assert_eq!(
-        build_operator_tree("true")
-            .unwrap()
-            .eval_boolean_with_context(&context),
-        Ok(true)
+        eval_string_with_context("3..3", &context),
+        Err(EvalexprError::VariableIdentifierNotFound("3..3".to_owned()))
     );
-    assert_eq!(
-        build_operator_tree("3,3").unwrap().eval_tuple(),
-        Ok(vec![Value::Int(3), Value::Int(3)])
-    );
-    assert_eq!(
-        build_operator_tree("3,3")
-            .unwrap()
-            .eval_tuple_with_context(&context),
-        Ok(vec![Value::Int(3), Value::Int(3)])
-    );
-    assert_eq!(
-        build_operator_tree("")
-            .unwrap()
-            .eval_empty_with_context(&context),
-        Ok(EMPTY_VALUE)
-    );
-
     assert_eq!(
         eval_string_with_context_mut("string", &mut context),
         Ok("a string".to_string())
     );
+    assert_eq!(
+        eval_string_with_context_mut("3.3", &mut context),
+        Err(EvalexprError::ExpectedString {
+            actual: Value::Float(3.3)
+        })
+    );
+    assert_eq!(
+        eval_string_with_context_mut("3..3", &mut context),
+        Err(EvalexprError::VariableIdentifierNotFound("3..3".to_owned()))
+    );
+
+    assert_eq!(eval_float("3.3"), Ok(3.3));
+    assert_eq!(
+        eval_float("33"),
+        Err(EvalexprError::ExpectedFloat {
+            actual: Value::Int(33)
+        })
+    );
+    assert_eq!(
+        eval_float("asd()"),
+        Err(EvalexprError::FunctionIdentifierNotFound("asd".to_owned()))
+    );
+    assert_eq!(eval_float_with_context("3.3", &context), Ok(3.3));
+    assert_eq!(
+        eval_float_with_context("33", &context),
+        Err(EvalexprError::ExpectedFloat {
+            actual: Value::Int(33)
+        })
+    );
+    assert_eq!(
+        eval_float_with_context("asd)", &context),
+        Err(EvalexprError::UnmatchedRBrace)
+    );
     assert_eq!(eval_float_with_context_mut("3.3", &mut context), Ok(3.3));
+    assert_eq!(
+        eval_float_with_context_mut("33", &mut context),
+        Err(EvalexprError::ExpectedFloat {
+            actual: Value::Int(33)
+        })
+    );
+    assert_eq!(
+        eval_float_with_context_mut("asd(", &mut context),
+        Err(EvalexprError::UnmatchedLBrace)
+    );
+
+    assert_eq!(eval_int("3"), Ok(3));
+    assert_eq!(
+        eval_int("3.3"),
+        Err(EvalexprError::ExpectedInt {
+            actual: Value::Float(3.3)
+        })
+    );
+    assert_eq!(
+        eval_int("(,);."),
+        Err(EvalexprError::VariableIdentifierNotFound(".".to_owned()))
+    );
+    assert_eq!(eval_int_with_context("3", &context), Ok(3));
+    assert_eq!(
+        eval_int_with_context("3.3", &context),
+        Err(EvalexprError::ExpectedInt {
+            actual: Value::Float(3.3)
+        })
+    );
+    assert_eq!(
+        eval_int_with_context("(,);.", &context),
+        Err(EvalexprError::VariableIdentifierNotFound(".".to_owned()))
+    );
     assert_eq!(eval_int_with_context_mut("3", &mut context), Ok(3));
+    assert_eq!(
+        eval_int_with_context_mut("3.3", &mut context),
+        Err(EvalexprError::ExpectedInt {
+            actual: Value::Float(3.3)
+        })
+    );
+    assert_eq!(
+        eval_int_with_context_mut("(,);.", &mut context),
+        Err(EvalexprError::VariableIdentifierNotFound(".".to_owned()))
+    );
+
+    assert_eq!(eval_number("3"), Ok(3.0));
+    assert_eq!(
+        eval_number("true"),
+        Err(EvalexprError::ExpectedNumber {
+            actual: Value::Boolean(true)
+        })
+    );
+    assert_eq!(
+        eval_number("abc"),
+        Err(EvalexprError::VariableIdentifierNotFound("abc".to_owned()))
+    );
+    assert_eq!(eval_number_with_context("3", &context), Ok(3.0));
+    assert_eq!(
+        eval_number_with_context("true", &context),
+        Err(EvalexprError::ExpectedNumber {
+            actual: Value::Boolean(true)
+        })
+    );
+    assert_eq!(
+        eval_number_with_context("abc", &context),
+        Err(EvalexprError::VariableIdentifierNotFound("abc".to_owned()))
+    );
     assert_eq!(eval_number_with_context_mut("3", &mut context), Ok(3.0));
+    assert_eq!(
+        eval_number_with_context_mut("true", &mut context),
+        Err(EvalexprError::ExpectedNumber {
+            actual: Value::Boolean(true)
+        })
+    );
+    assert_eq!(
+        eval_number_with_context_mut("abc", &mut context),
+        Err(EvalexprError::VariableIdentifierNotFound("abc".to_owned()))
+    );
+
+    assert_eq!(eval_boolean("true"), Ok(true));
+    assert_eq!(
+        eval_boolean("4"),
+        Err(EvalexprError::ExpectedBoolean {
+            actual: Value::Int(4)
+        })
+    );
+    assert_eq!(
+        eval_boolean("trueee"),
+        Err(EvalexprError::VariableIdentifierNotFound(
+            "trueee".to_owned()
+        ))
+    );
+    assert_eq!(eval_boolean_with_context("true", &context), Ok(true));
+    assert_eq!(
+        eval_boolean_with_context("4", &context),
+        Err(EvalexprError::ExpectedBoolean {
+            actual: Value::Int(4)
+        })
+    );
+    assert_eq!(
+        eval_boolean_with_context("trueee", &context),
+        Err(EvalexprError::VariableIdentifierNotFound(
+            "trueee".to_owned()
+        ))
+    );
     assert_eq!(
         eval_boolean_with_context_mut("true", &mut context),
         Ok(true)
     );
+    assert_eq!(
+        eval_boolean_with_context_mut("4", &mut context),
+        Err(EvalexprError::ExpectedBoolean {
+            actual: Value::Int(4)
+        })
+    );
+    assert_eq!(
+        eval_boolean_with_context_mut("trueee", &mut context),
+        Err(EvalexprError::VariableIdentifierNotFound(
+            "trueee".to_owned()
+        ))
+    );
+
+    assert_eq!(eval_tuple("3,3"), Ok(vec![Value::Int(3), Value::Int(3)]));
+    assert_eq!(
+        eval_tuple("33"),
+        Err(EvalexprError::ExpectedTuple {
+            actual: Value::Int(33)
+        })
+    );
+    assert_eq!(
+        eval_tuple("3a3"),
+        Err(EvalexprError::VariableIdentifierNotFound("3a3".to_owned()))
+    );
+    assert_eq!(
+        eval_tuple_with_context("3,3", &context),
+        Ok(vec![Value::Int(3), Value::Int(3)])
+    );
+    assert_eq!(
+        eval_tuple_with_context("33", &context),
+        Err(EvalexprError::ExpectedTuple {
+            actual: Value::Int(33)
+        })
+    );
+    assert_eq!(
+        eval_tuple_with_context("3a3", &context),
+        Err(EvalexprError::VariableIdentifierNotFound("3a3".to_owned()))
+    );
     assert_eq!(
         eval_tuple_with_context_mut("3,3", &mut context),
         Ok(vec![Value::Int(3), Value::Int(3)])
     );
+    assert_eq!(
+        eval_tuple_with_context_mut("33", &mut context),
+        Err(EvalexprError::ExpectedTuple {
+            actual: Value::Int(33)
+        })
+    );
+    assert_eq!(
+        eval_tuple_with_context_mut("3a3", &mut context),
+        Err(EvalexprError::VariableIdentifierNotFound("3a3".to_owned()))
+    );
+
+    assert_eq!(eval_empty(""), Ok(EMPTY_VALUE));
+    assert_eq!(eval_empty("()"), Ok(EMPTY_VALUE));
+    assert_eq!(
+        eval_empty("(,)"),
+        Err(EvalexprError::ExpectedEmpty {
+            actual: Value::Tuple(vec![Value::Empty, Value::Empty])
+        })
+    );
+    assert_eq!(
+        eval_empty("xaq"),
+        Err(EvalexprError::VariableIdentifierNotFound("xaq".to_owned()))
+    );
+    assert_eq!(eval_empty_with_context("", &context), Ok(EMPTY_VALUE));
+    assert_eq!(eval_empty_with_context("()", &context), Ok(EMPTY_VALUE));
+    assert_eq!(
+        eval_empty_with_context("(,)", &context),
+        Err(EvalexprError::ExpectedEmpty {
+            actual: Value::Tuple(vec![Value::Empty, Value::Empty])
+        })
+    );
+    assert_eq!(
+        eval_empty_with_context("xaq", &context),
+        Err(EvalexprError::VariableIdentifierNotFound("xaq".to_owned()))
+    );
     assert_eq!(
         eval_empty_with_context_mut("", &mut context),
         Ok(EMPTY_VALUE)
     );
+    assert_eq!(
+        eval_empty_with_context_mut("()", &mut context),
+        Ok(EMPTY_VALUE)
+    );
+    assert_eq!(
+        eval_empty_with_context_mut("(,)", &mut context),
+        Err(EvalexprError::ExpectedEmpty {
+            actual: Value::Tuple(vec![Value::Empty, Value::Empty])
+        })
+    );
+    assert_eq!(
+        eval_empty_with_context_mut("xaq", &mut context),
+        Err(EvalexprError::VariableIdentifierNotFound("xaq".to_owned()))
+    );
 
+    // With detour via build_operator_tree
+
+    assert_eq!(
+        build_operator_tree("\"3.3\"").unwrap().eval_string(),
+        Ok("3.3".to_owned())
+    );
+    assert_eq!(
+        build_operator_tree("3.3").unwrap().eval_string(),
+        Err(EvalexprError::ExpectedString {
+            actual: Value::Float(3.3)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("3..3").unwrap().eval_string(),
+        Err(EvalexprError::VariableIdentifierNotFound("3..3".to_owned()))
+    );
+    assert_eq!(
+        build_operator_tree("string")
+            .unwrap()
+            .eval_string_with_context(&context),
+        Ok("a string".to_owned())
+    );
+    assert_eq!(
+        build_operator_tree("3.3")
+            .unwrap()
+            .eval_string_with_context(&context),
+        Err(EvalexprError::ExpectedString {
+            actual: Value::Float(3.3)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("3..3")
+            .unwrap()
+            .eval_string_with_context(&context),
+        Err(EvalexprError::VariableIdentifierNotFound("3..3".to_owned()))
+    );
     assert_eq!(
         build_operator_tree("string")
             .unwrap()
             .eval_string_with_context_mut(&mut context),
         Ok("a string".to_string())
     );
+    assert_eq!(
+        build_operator_tree("3.3")
+            .unwrap()
+            .eval_string_with_context_mut(&mut context),
+        Err(EvalexprError::ExpectedString {
+            actual: Value::Float(3.3)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("3..3")
+            .unwrap()
+            .eval_string_with_context_mut(&mut context),
+        Err(EvalexprError::VariableIdentifierNotFound("3..3".to_owned()))
+    );
+
+    assert_eq!(build_operator_tree("3.3").unwrap().eval_float(), Ok(3.3));
+    assert_eq!(
+        build_operator_tree("33").unwrap().eval_float(),
+        Err(EvalexprError::ExpectedFloat {
+            actual: Value::Int(33)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("asd()").unwrap().eval_float(),
+        Err(EvalexprError::FunctionIdentifierNotFound("asd".to_owned()))
+    );
+    assert_eq!(
+        build_operator_tree("3.3")
+            .unwrap()
+            .eval_float_with_context(&context),
+        Ok(3.3)
+    );
+    assert_eq!(
+        build_operator_tree("33")
+            .unwrap()
+            .eval_float_with_context(&context),
+        Err(EvalexprError::ExpectedFloat {
+            actual: Value::Int(33)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("asd")
+            .unwrap()
+            .eval_float_with_context(&context),
+        Err(EvalexprError::VariableIdentifierNotFound("asd".to_owned()))
+    );
     assert_eq!(
         build_operator_tree("3.3")
             .unwrap()
             .eval_float_with_context_mut(&mut context),
         Ok(3.3)
     );
+    assert_eq!(
+        build_operator_tree("33")
+            .unwrap()
+            .eval_float_with_context_mut(&mut context),
+        Err(EvalexprError::ExpectedFloat {
+            actual: Value::Int(33)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("asd")
+            .unwrap()
+            .eval_float_with_context_mut(&mut context),
+        Err(EvalexprError::VariableIdentifierNotFound("asd".to_owned()))
+    );
+
+    assert_eq!(build_operator_tree("3").unwrap().eval_int(), Ok(3));
+    assert_eq!(
+        build_operator_tree("3.3").unwrap().eval_int(),
+        Err(EvalexprError::ExpectedInt {
+            actual: Value::Float(3.3)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("(,);.").unwrap().eval_int(),
+        Err(EvalexprError::VariableIdentifierNotFound(".".to_owned()))
+    );
+    assert_eq!(
+        build_operator_tree("3")
+            .unwrap()
+            .eval_int_with_context(&context),
+        Ok(3)
+    );
+    assert_eq!(
+        build_operator_tree("3.3")
+            .unwrap()
+            .eval_int_with_context(&context),
+        Err(EvalexprError::ExpectedInt {
+            actual: Value::Float(3.3)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("(,);.")
+            .unwrap()
+            .eval_int_with_context(&context),
+        Err(EvalexprError::VariableIdentifierNotFound(".".to_owned()))
+    );
     assert_eq!(
         build_operator_tree("3")
             .unwrap()
             .eval_int_with_context_mut(&mut context),
         Ok(3)
     );
+    assert_eq!(
+        build_operator_tree("3.3")
+            .unwrap()
+            .eval_int_with_context_mut(&mut context),
+        Err(EvalexprError::ExpectedInt {
+            actual: Value::Float(3.3)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("(,);.")
+            .unwrap()
+            .eval_int_with_context_mut(&mut context),
+        Err(EvalexprError::VariableIdentifierNotFound(".".to_owned()))
+    );
+
+    assert_eq!(build_operator_tree("3").unwrap().eval_number(), Ok(3.0));
+    assert_eq!(
+        build_operator_tree("true").unwrap().eval_number(),
+        Err(EvalexprError::ExpectedNumber {
+            actual: Value::Boolean(true)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("abc").unwrap().eval_number(),
+        Err(EvalexprError::VariableIdentifierNotFound("abc".to_owned()))
+    );
+    assert_eq!(
+        build_operator_tree("3")
+            .unwrap()
+            .eval_number_with_context(&context),
+        Ok(3.0)
+    );
+    assert_eq!(
+        build_operator_tree("true")
+            .unwrap()
+            .eval_number_with_context(&context),
+        Err(EvalexprError::ExpectedNumber {
+            actual: Value::Boolean(true)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("abc")
+            .unwrap()
+            .eval_number_with_context(&context),
+        Err(EvalexprError::VariableIdentifierNotFound("abc".to_owned()))
+    );
     assert_eq!(
         build_operator_tree("3")
             .unwrap()
             .eval_number_with_context_mut(&mut context),
         Ok(3.0)
     );
+    assert_eq!(
+        build_operator_tree("true")
+            .unwrap()
+            .eval_number_with_context_mut(&mut context),
+        Err(EvalexprError::ExpectedNumber {
+            actual: Value::Boolean(true)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("abc")
+            .unwrap()
+            .eval_number_with_context_mut(&mut context),
+        Err(EvalexprError::VariableIdentifierNotFound("abc".to_owned()))
+    );
+
+    assert_eq!(
+        build_operator_tree("true").unwrap().eval_boolean(),
+        Ok(true)
+    );
+    assert_eq!(
+        build_operator_tree("4").unwrap().eval_boolean(),
+        Err(EvalexprError::ExpectedBoolean {
+            actual: Value::Int(4)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("trueee").unwrap().eval_boolean(),
+        Err(EvalexprError::VariableIdentifierNotFound(
+            "trueee".to_owned()
+        ))
+    );
+    assert_eq!(
+        build_operator_tree("true")
+            .unwrap()
+            .eval_boolean_with_context(&context),
+        Ok(true)
+    );
+    assert_eq!(
+        build_operator_tree("4")
+            .unwrap()
+            .eval_boolean_with_context(&context),
+        Err(EvalexprError::ExpectedBoolean {
+            actual: Value::Int(4)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("trueee")
+            .unwrap()
+            .eval_boolean_with_context(&context),
+        Err(EvalexprError::VariableIdentifierNotFound(
+            "trueee".to_owned()
+        ))
+    );
     assert_eq!(
         build_operator_tree("true")
             .unwrap()
             .eval_boolean_with_context_mut(&mut context),
         Ok(true)
     );
+    assert_eq!(
+        build_operator_tree("4")
+            .unwrap()
+            .eval_boolean_with_context_mut(&mut context),
+        Err(EvalexprError::ExpectedBoolean {
+            actual: Value::Int(4)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("trueee")
+            .unwrap()
+            .eval_boolean_with_context_mut(&mut context),
+        Err(EvalexprError::VariableIdentifierNotFound(
+            "trueee".to_owned()
+        ))
+    );
+
+    assert_eq!(
+        build_operator_tree("3,3").unwrap().eval_tuple(),
+        Ok(vec![Value::Int(3), Value::Int(3)])
+    );
+    assert_eq!(
+        build_operator_tree("33").unwrap().eval_tuple(),
+        Err(EvalexprError::ExpectedTuple {
+            actual: Value::Int(33)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("3a3").unwrap().eval_tuple(),
+        Err(EvalexprError::VariableIdentifierNotFound("3a3".to_owned()))
+    );
+    assert_eq!(
+        build_operator_tree("3,3")
+            .unwrap()
+            .eval_tuple_with_context(&context),
+        Ok(vec![Value::Int(3), Value::Int(3)])
+    );
+    assert_eq!(
+        build_operator_tree("33")
+            .unwrap()
+            .eval_tuple_with_context(&context),
+        Err(EvalexprError::ExpectedTuple {
+            actual: Value::Int(33)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("3a3")
+            .unwrap()
+            .eval_tuple_with_context(&context),
+        Err(EvalexprError::VariableIdentifierNotFound("3a3".to_owned()))
+    );
     assert_eq!(
         build_operator_tree("3,3")
             .unwrap()
             .eval_tuple_with_context_mut(&mut context),
         Ok(vec![Value::Int(3), Value::Int(3)])
     );
+    assert_eq!(
+        build_operator_tree("33")
+            .unwrap()
+            .eval_tuple_with_context_mut(&mut context),
+        Err(EvalexprError::ExpectedTuple {
+            actual: Value::Int(33)
+        })
+    );
+    assert_eq!(
+        build_operator_tree("3a3")
+            .unwrap()
+            .eval_tuple_with_context_mut(&mut context),
+        Err(EvalexprError::VariableIdentifierNotFound("3a3".to_owned()))
+    );
+
+    assert_eq!(
+        build_operator_tree("").unwrap().eval_empty(),
+        Ok(EMPTY_VALUE)
+    );
+    assert_eq!(
+        build_operator_tree("()").unwrap().eval_empty(),
+        Ok(EMPTY_VALUE)
+    );
+    assert_eq!(
+        build_operator_tree("(,)").unwrap().eval_empty(),
+        Err(EvalexprError::ExpectedEmpty {
+            actual: Value::Tuple(vec![Value::Empty, Value::Empty])
+        })
+    );
+    assert_eq!(
+        build_operator_tree("xaq").unwrap().eval_empty(),
+        Err(EvalexprError::VariableIdentifierNotFound("xaq".to_owned()))
+    );
+    assert_eq!(
+        build_operator_tree("")
+            .unwrap()
+            .eval_empty_with_context(&context),
+        Ok(EMPTY_VALUE)
+    );
+    assert_eq!(
+        build_operator_tree("()")
+            .unwrap()
+            .eval_empty_with_context(&context),
+        Ok(EMPTY_VALUE)
+    );
+    assert_eq!(
+        build_operator_tree("(,)")
+            .unwrap()
+            .eval_empty_with_context(&context),
+        Err(EvalexprError::ExpectedEmpty {
+            actual: Value::Tuple(vec![Value::Empty, Value::Empty])
+        })
+    );
+    assert_eq!(
+        build_operator_tree("xaq")
+            .unwrap()
+            .eval_empty_with_context(&context),
+        Err(EvalexprError::VariableIdentifierNotFound("xaq".to_owned()))
+    );
     assert_eq!(
         build_operator_tree("")
             .unwrap()
             .eval_empty_with_context_mut(&mut context),
         Ok(EMPTY_VALUE)
     );
+    assert_eq!(
+        build_operator_tree("()")
+            .unwrap()
+            .eval_empty_with_context_mut(&mut context),
+        Ok(EMPTY_VALUE)
+    );
+    assert_eq!(
+        build_operator_tree("(,)")
+            .unwrap()
+            .eval_empty_with_context_mut(&mut context),
+        Err(EvalexprError::ExpectedEmpty {
+            actual: Value::Tuple(vec![Value::Empty, Value::Empty])
+        })
+    );
+    assert_eq!(
+        build_operator_tree("xaq")
+            .unwrap()
+            .eval_empty_with_context_mut(&mut context),
+        Err(EvalexprError::VariableIdentifierNotFound("xaq".to_owned()))
+    );
 }
 
 #[test]
@@ -603,12 +1193,24 @@ fn test_strings() {
     assert_eq!(eval("\"a\" + \"b\""), Ok(Value::from("ab")));
     assert_eq!(eval("\"a\" > \"b\""), Ok(Value::from(false)));
     assert_eq!(eval("\"a\" < \"b\""), Ok(Value::from(true)));
+    assert_eq!(eval("\"a\" >= \"b\""), Ok(Value::from(false)));
+    assert_eq!(eval("\"a\" <= \"b\""), Ok(Value::from(true)));
+    assert_eq!(eval("\"a\" >= \"a\""), Ok(Value::from(true)));
+    assert_eq!(eval("\"a\" <= \"a\""), Ok(Value::from(true)));
     assert_eq!(eval("\"xa\" > \"xb\""), Ok(Value::from(false)));
     assert_eq!(eval("\"xa\" < \"xb\""), Ok(Value::from(true)));
     assert_eq!(eval("\"{}\" != \"{}\""), Ok(Value::from(false)));
     assert_eq!(eval("\"{}\" == \"{}\""), Ok(Value::from(true)));
 }
 
+#[test]
+fn test_string_escaping() {
+    assert_eq!(
+        eval("\"\\\"str\\\\ing\\\"\""),
+        Ok(Value::from("\"str\\ing\""))
+    );
+}
+
 #[test]
 fn test_tuple_definitions() {
     assert_eq!(eval_empty("()"), Ok(()));
@@ -909,3 +1511,61 @@ fn test_error_constructors() {
         })
     );
 }
+
+#[test]
+fn test_iterators() {
+    let tree = build_operator_tree("5 + 3 + fun(4) + var").unwrap();
+    let mut iter = tree.iter_identifiers();
+    assert_eq!(iter.next(), Some("fun"));
+    assert_eq!(iter.next(), Some("var"));
+    assert_eq!(iter.next(), None);
+
+    let mut iter = tree.iter_variable_identifiers();
+    assert_eq!(iter.next(), Some("var"));
+    assert_eq!(iter.next(), None);
+
+    let mut iter = tree.iter_function_identifiers();
+    assert_eq!(iter.next(), Some("fun"));
+    assert_eq!(iter.next(), None);
+}
+
+#[test]
+fn test_same_operator_chains() {
+    #![allow(clippy::eq_op)]
+    assert_eq!(
+        eval("3.0 / 3.0 / 3.0 / 3.0"),
+        Ok(Value::from(3.0 / 3.0 / 3.0 / 3.0))
+    );
+    assert_eq!(
+        eval("3.0 - 3.0 - 3.0 - 3.0"),
+        Ok(Value::from(3.0 - 3.0 - 3.0 - 3.0))
+    );
+}
+
+#[test]
+fn test_long_expression_i89() {
+    let tree = build_operator_tree(
+        "x*0.2*5/4+x*2*4*1*1*1*1*1*1*1+7*math::sin(y)-z/math::sin(3.0/2.0/(1-x*4*1*1*1*1))",
+    )
+    .unwrap();
+    let x = 0.0;
+    let y: f64 = 3.0;
+    let z = 4.0;
+    let context = context_map! {
+        "x" => 0.0,
+        "y" => 3.0,
+        "z" => 4.0
+    }
+    .unwrap();
+    let expected = x * 0.2 * 5.0 / 4.0
+        + x * 2.0 * 4.0 * 1.0 * 1.0 * 1.0 * 1.0 * 1.0 * 1.0 * 1.0
+        + 7.0 * y.sin()
+        - z / (3.0 / 2.0 / (1.0 - x * 4.0 * 1.0 * 1.0 * 1.0 * 1.0)).sin();
+    let actual = tree.eval_float_with_context(&context).unwrap();
+    assert!(
+        (expected - actual).abs() < expected.abs().min(actual.abs()) * 1e-12,
+        "expected: {}, actual: {}",
+        expected,
+        actual
+    );
+}
